import argparse
import curses
import locale
import logging
import os
import sys
import time

import game
import screen
import userinput


class GameWorld(object):
    def __init__(self, game_cls):
        self.game_cls = game_cls

    def _main(self, stdscr, args):
        # Initialize screen.
        scr = screen.Screen(stdscr,
                            width=getattr(self.game_cls, 'SCREEN_WIDTH', None),
                            height=getattr(self.game_cls, 'SCREEN_HEIGHT', None),
                            show_fps=args.show_fps)

        # Initialize user input.
        cur_userinput = userinput.UserInput(scr)

        # Initialize game.
        cur_game = self.game_cls(scr, cur_userinput, args)

        last_tick = time.time() * 1000
        while True:
            tick_length = cur_game.get_tick_length_ms()
            time_since_last_tick = time.time() * 1000 - last_tick
            ticks_elapsed = int(time_since_last_tick / tick_length)
            last_tick += ticks_elapsed * tick_length
            cur_game.iterate(ticks_elapsed)

            cmd = cur_userinput.getcmd()
            cur_game.handle_input(cmd)

            if args.show_last_char:
                s = 'last_chr:{}'.format(cur_userinput.last_char)
                scr.stdscr.addstr(2, scr.width-len(s), s)

            if not cur_game._running:
                break

            if not cur_game.only_draw_on_ticks or ticks_elapsed:
                scr.draw()

        cur_game.quit()
        return 0

    def main(self, args):
        # Initialize UTF-8.
        locale.setlocale(locale.LC_ALL, '')

        # Reduce the escape key delay.
        os.environ['ESCDELAY'] = '5'

        # Parse and store command-line args.
        arg_parser = argparse.ArgumentParser(description='The GameWorld engine.')

        if hasattr(self.game_cls, 'add_args'):
            self.game_cls.add_args(arg_parser)

        arg_parser.add_argument(
            '--log-level',
            help='The lowest level of messages to log to a file.',
            default='INFO',
            choices=['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG'])
        arg_parser.add_argument(
            '--log-filename',
            help='The filename of the log file.',
            default='gameworld.log')
        arg_parser.add_argument(
            '--show-fps',
            help='Show the renderer\'s frames per second on the display.',
            action='store_true')
        arg_parser.add_argument(
            '--show-last-char',
            help='Show the value of the last key entered.',
            action='store_true')
        args = arg_parser.parse_args(args)

        # Initialize logging.
        logging.basicConfig(filename=args.log_filename, level=args.log_level)

        return curses.wrapper(self._main, args)
