import curses
import logging
import pickle
import time


MAX_COLORS = 8


def load_figure(filename):
    with open(filename, 'rb') as fh:
        return pickle.load(fh)


class Figure(object):
    def __init__(self, width, height, x_pos=0, y_pos=0, z_index=1, *args, **kwargs):
        self.width = width
        self.height = height
        self.x_pos = x_pos
        self.y_pos = y_pos
        self.z_index = z_index
        self.visible = True

    def set_visible(self, visible=True):
        self.visible = visible

    def get_char_and_attr(self, x, y):
        """Returns the character and attribute value for the fig at (x,y).

        Args x and y are offsets from the figure's upper-left corner,
        where 0,0 is the character in the first row, first column of
        the figure.

        If the figure has no character at that offset and wishes to be
        transparent, this method returns None for the char value (and
        attr is ignored).

        Args:
          x: The x offset from the left-most column of the figure.
          y: The y offset from the top-most row of the figure.

        Returns:
          (char, attr) where char is the ordinal character value and
          attr is the curses attribute value (the bitwise-or of
          various flags, 0 for all defaults).  If the figure is
          transparent at the given location, char is None.
        """
        raise NotImplemented

    def __str__(self):
        return ('Figure(width={}, height={}, x_pos={}, y_pos={}, '
                'z_index={}, visible={})'
                .format(self.width, self.height, self.x_pos, self.y_pos,
                        self.z_index, self.visible))


class DrawnFigure(Figure):
    def __init__(self, *args, **kwargs):
        self.frame_count = kwargs.get('frame_count', 1)
        if 'frame_count' in kwargs:
            del kwargs['frame_count']
        super(DrawnFigure, self).__init__(*args, **kwargs)

        self.char_array = [32] * (self.width * self.height * self.frame_count)
        self.color_array = [0] * (self.width * self.height * self.frame_count)
        self.cur_frame = 0

    def clear_frame(self):
        frame_start_i = self._get_index_for_coordinate(0, 0, frame)
        frame_size = self.width * self.height
        for i in frame_size:
            self.char_array[frame_start_i + i] = 32
            self.color_array[frame_start_i + i] = 0

    def _get_index_for_coordinate(self, x, y, frame=None):
        if frame is None:
            frame = self.cur_frame
        i = frame * self.width * self.height + x * self.height + y
        return i

    def get_char_and_attr(self, x, y, frame=None):
        i = self._get_index_for_coordinate(x, y, frame=frame)
        if self.char_array[i] == -1:
            return (None, None)
        return self.char_array[i], self.color_array[i]

    def set_char_and_attr(self, x, y, char, attr, frame=None):
        """
        Args:
          x:
          y:
          char: ord for the char, -1 for transparent, or None to not change.
          attr: curses attrs, or None to not change.
          frame:
        """
        i = self._get_index_for_coordinate(x, y, frame=frame)
        if char is not None:
            self.char_array[i] = char
        if attr is not None:
            self.color_array[i] = attr

    def write_str(self, x, y, s, attr, frame=None):
        assert x + len(s) < self.width
        for i, c in enumerate(s):
            self.set_char_and_attr(x+i, y, ord(c), attr, frame=frame)

    def duplicate_frame(self, frame=None):
        if frame is None:
            frame = self.cur_frame
        frame_start_i = self._get_index_for_coordinate(0, 0, frame)
        new_frame_start_i = self._get_index_for_coordinate(0, 0, frame + 1)
        frame_size = self.width * self.height
        for i in xrange(frame_size):
            self.char_array.insert(new_frame_start_i + i,
                                   self.char_array[frame_start_i + i])
            self.color_array.insert(new_frame_start_i + i,
                                    self.color_array[frame_start_i + i])
        self.frame_count += 1

    def delete_frame(self, frame=None):
        if self.frame_count == 1:
            return
        if frame is None:
            frame = self.cur_frame
        frame_start_i = self._get_index_for_coordinate(0, 0, frame)
        frame_size = self.width * self.height
        del self.char_array[frame_start_i:frame_size]
        del self.color_array[frame_start_i:frame_size]
        self.frame_count -= 1
        if self.cur_frame >= self.frame_count:
            self.cur_frame -= 1

    def clear_frame(self, frame=None):
        frame_start_i = self._get_index_for_coordinate(0, 0, frame=frame)
        frame_size = self.width * self.height
        for i in xrange(frame_size):
            self.char_array[frame_start_i + i] = 32
            self.color_array[frame_start_i + i] = 0

    # def resize(self, new_width, new_height, x_offset, y_offset):


class Screen(object):
    def __init__(self, stdscr, width=None, height=None, show_fps=False):
        self.stdscr = stdscr
        self.show_fps = show_fps
        self.width = width or stdscr.getmaxyx()[1]
        self.height = height or stdscr.getmaxyx()[0]
        self._show_fps = show_fps
        self._init_time = time.time()
        self._draw_count = 0
        self._figures = []

        # Initialize curses.
        curses.curs_set(0)
        stdscr.nodelay(True)
        self.win = curses.newwin(self.height, self.width, 0, 0)
        self.win.scrollok(False)

        # Initialize color pairs.
        # Color pair 0 is white-on-black and cannot be changed.
        # TODO: init more color pairs for bg colors?
        for n in range(1, 8):
            curses.init_pair(n, n, curses.COLOR_BLACK)

    def add_figure(self, figure):
        # Add the figure to the list, in reverse z_index order (highest first).
        for i, f in enumerate(self._figures):
            if f.z_index < figure.z_index:
                self._figures.insert(i, figure)
                return
        self._figures.append(figure)

    def draw2(self):
        """Draw a frame.

        This implementation paints each figure onto the screen, in
        z-index order.  This favors redrawing the same location
        multiple times in the case of overlap, in favor of only
        examining each figure once.
        """
        self.win.clear()

        for f_i in xrange(len(self._figures) - 1, -1, -1):
            f = self._figures[f_i]
            if f.visible:
                for j in range(f.height):
                    for i in range(f.width):
                        char_x = f.x_pos + i
                        char_y = f.y_pos + j
                        (ch, attr) = f.get_char_and_attr(i, j)
                        if ((0 <= char_x < self.width) and
                            (0 <= char_y < self.height) and
                            (ch is not None)):
                            # TODO: handle lower right corner...
                            self.win.addch(char_y, char_x, ch, attr)

        if self._show_fps:
            # TODO: use running average instead of overall average
            self._draw_count += 1
            fps = self._draw_count / (time.time() - self._init_time)
            fps_str = ' FPS: {:.2f}'.format(fps)
            self.win.addstr(0, self.width - len(fps_str) - 1, fps_str)

        self.win.refresh()        

    def draw(self):
        """Draw a frame.

        This implementation iterates over every screen position and
        scans the figure list to calculate the character to draw.
        This minimizes calls to addch(), at the expense of examining
        figures at every location.

        See draw2() above for an alternative (heed TODOs), though I'm
        not currently hindered by the speed of updating the curses
        buffer so I'm going to leave it as is for now.
        """
        # Clear the curses window buffer.
        self.win.clear()

        # If the user's terminal is less than height,width, print a msg
        (max_y, max_x) = self.win.getmaxyx()
        if max_y < self.height or max_x < self.width:
            self.win.addstr(0, 0,
                            'Please adjust your terminal to {}x{}.'
                            .format(self.width, self.height))
            self.win.refresh()
            return

        # Write all of the window characters to the curses window buffer.
        #
        # The character in the lower right corner needs special
        # handling because it cannot be drawn with addch() without
        # scrolling the window (which is an error if the window does
        # not scroll). We remember the second-to-last character, then
        # when in the corner, we addch the corner character (if any)
        # one space over, then insch the second-to-last character (or
        # a space if none) into place.

        lower_right_one_over = None
        for j in range(self.height):
            for i in range(self.width):
                for f in self._figures:
                    if (f.visible and
                        (f.x_pos <= i < f.x_pos + f.width) and
                        (f.y_pos <= j < f.y_pos + f.height)):
                        (ch, attr) = f.get_char_and_attr(i-f.x_pos,
                                                         j-f.y_pos)
                        if ch is not None:
                            if (j == self.height - 1 and i >= self.width - 2):
                                if i == self.width - 2:
                                    # There is a second-to-last character.
                                    # Remember it but don't draw it.
                                    lower_right_one_over = (ch, attr)
                                else:
                                    # There is a corner character.
                                    # Draw it in the second-to-last location.
                                    self.win.addch(j, i-1, ch, attr)
                            else:
                                self.win.addch(j, i, ch, attr)
                            break  # figure loop

                if (j == self.height - 1 and
                    i == self.width - 1):
                    # This is the last loop.
                    if lower_right_one_over is not None:
                        # There was a second-to-last character.
                        # Insert it, pushing the corner into place.
                        self.win.insch(j, i-1,
                                       lower_right_one_over[0],
                                       lower_right_one_over[1])
                    else:
                        # There was not a second-to-last character.
                        # Insert a space in case there was a corner character.
                        self.win.insch(j, i-1, ord(' '))

        if self._show_fps:
            # TODO: use running average instead of overall average
            self._draw_count += 1
            fps = self._draw_count / (time.time() - self._init_time)
            fps_str = ' FPS: {:.2f}'.format(fps)
            self.win.addstr(0, self.width - len(fps_str) - 1, fps_str)

        # Copy the window buffer to the screen.
        self.win.refresh()
